@@ -4,23 +4,26 @@ import logging
 import itertools
 import contextlib
 import dataclasses
-import async_timeout
-
 from collections import Counter, defaultdict
 
+import async_timeout
+
 import zigpy_znp.types as t
 import zigpy_znp.config as conf
+import zigpy_znp.logger as log
 import zigpy_znp.commands as c
-
-from zigpy_znp.types import nvids
-from zigpy_znp.logger import TRACE
-
 from zigpy_znp import uart
+from zigpy_znp.types import nvids
 from zigpy_znp.frames import GeneralFrame
-from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse
-
+from zigpy_znp.exceptions import (
+    SecurityError,
+    CommandNotRecognized,
+    InvalidCommandResponse,
+)
 
 LOGGER = logging.getLogger(__name__)
+AFTER_CONNECT_DELAY = 1  # seconds
+STARTUP_DELAY = 1  # seconds
 
 
 def _deduplicate_commands(
@@ -169,6 +172,8 @@ class ZNP:
         self._uart = None
         self._app = None
         self._config = config
+        self._version = None
+        self._capabilities = t.MTCapabilities(0)
 
         self._listeners = defaultdict(list)
         self._sync_request_lock = asyncio.Lock()
@@ -195,40 +200,69 @@ class ZNP:
         try:
             self._uart = await uart.connect(self._config[conf.CONF_DEVICE], self)
 
-            if self._config[conf.CONF_ZNP_CONFIG][conf.CONF_SKIP_BOOTLOADER]:
-                LOGGER.debug("Sending special byte to skip the bootloader")
-                self._uart._transport_write(bytes([c.ubl.BootloaderRunMode.FORCE_RUN]))
+            LOGGER.debug("Waiting %ss before sending anything", AFTER_CONNECT_DELAY)
+            await asyncio.sleep(AFTER_CONNECT_DELAY)
 
-            # We have to disable all non-bootloader commands to enter the
+            if self._config[conf.CONF_ZNP_CONFIG][conf.CONF_SKIP_BOOTLOADER]:
+                LOGGER.debug("Sending bootloader skip byte")
+
+                # XXX: Z-Stack locks up if other radios try probing it first.
+                #      Writing the bootloader skip byte a bunch of times (at least 167)
+                #      appears to reset it.
+                skip = bytes([c.ubl.BootloaderRunMode.FORCE_RUN])
+                self._uart._transport_write(skip * 256)
+
+            # We have to disable all non-bootloader commands to enter the serial
             # bootloader upon connecting to the UART.
             if test_port:
+                # Some Z-Stack 3 devices don't like you sending data immediately after
+                # opening the serial port. A small delay helps, but they also sometimes
+                # send a reset indication message when they're ready.
                 LOGGER.debug(
-                    "Testing connection to %s", self._uart.transport.serial.name
+                    "Waiting %ss or until a reset indication is received", STARTUP_DELAY
                 )
 
-                # Make sure that our port works
-                ping_rsp = await self.request(c.SYS.Ping.Req())
+                try:
+                    async with async_timeout.timeout(STARTUP_DELAY):
+                        await self.wait_for_response(
+                            c.SYS.ResetInd.Callback(partial=True)
+                        )
+                except asyncio.TimeoutError:
+                    pass
 
-                if not ping_rsp.Capabilities & t.MTCapabilities.CAP_APP_CNF:
-                    old_version_msg = (
-                        "Your device appears to be running an old version of Z-Stack."
-                        " The earliest supported release is Z-Stack 3.0.1."
+                LOGGER.debug("Testing connection to %s", self._port_path)
+
+                # Make sure that our port works
+                version = await self.request(c.SYS.Version.Req())
+
+                if (version.TransportRev, version.MajorRel, version.MinorRel) not in {
+                    (2, 2, 6),
+                    (2, 2, 7),
+                }:
+                    raise RuntimeError(
+                        f"Your device is running an unsupported"
+                        f" release of Z-Stack: {version}"
                     )
 
-                    if check_version:
-                        raise RuntimeError(old_version_msg)
-                    else:
-                        LOGGER.warning(old_version_msg)
+                self._version = version
+                self._capabilities = (await self.request(c.SYS.Ping.Req())).Capabilities
         except Exception:
+            LOGGER.debug("Connection to %s failed, cleaning up", self._port_path)
             self.close()
             raise
 
         LOGGER.debug(
             "Connected to %s at %s baud",
-            self._uart.transport.serial.name,
-            self._uart.transport.serial.baudrate,
+            self._uart._transport.serial.name,
+            self._uart._transport.serial.baudrate,
         )
 
+    def connection_made(self) -> None:
+        """
+        Called by the UART object when a connection has been made.
+        """
+        pass
+
     def connection_lost(self, exc) -> None:
         """
         Called by the UART object to indicate that the port was closed. Propagates up
@@ -237,14 +271,14 @@ class ZNP:
 
         LOGGER.debug("We were disconnected from %s: %s", self._port_path, exc)
 
-        # The UART is already closed, there is no point in closing it again
+        # The UART is already closed by the time it notifies us
         self._uart = None
 
+        self.close()
+
         if self._app is not None:
             self._app.connection_lost(exc)
 
-        self.close()
-
     def close(self) -> None:
         """
         Cleans up resources, namely the listener queues.
@@ -257,13 +291,12 @@ class ZNP:
                 listener.cancel()
 
         self._listeners.clear()
+        self._version = None
 
         if self._uart is not None:
             self._uart.close()
             self._uart = None
 
-        self._app = None
-
     def remove_listener(self, listener: BaseResponseListener) -> None:
         """
         Unbinds a listener from ZNP.
@@ -276,14 +309,17 @@ class ZNP:
         if not self._listeners:
             return
 
-        LOGGER.log(TRACE, "Removing listener %s", listener)
+        LOGGER.log(log.TRACE, "Removing listener %s", listener)
 
         for header in listener.matching_headers():
-            self._listeners[header].remove(listener)
+            try:
+                self._listeners[header].remove(listener)
+            except ValueError:
+                pass
 
             if not self._listeners[header]:
                 LOGGER.log(
-                    TRACE, "Cleaning up empty listener list for header %s", header
+                    log.TRACE, "Cleaning up empty listener list for header %s", header
                 )
                 del self._listeners[header]
 
@@ -292,15 +328,17 @@ class ZNP:
         for listener in itertools.chain.from_iterable(self._listeners.values()):
             counts[type(listener)] += 1
 
-        LOGGER.debug(
+        LOGGER.log(
+            log.TRACE,
             "There are %d callbacks and %d one-shot listeners remaining",
             counts[CallbackResponseListener],
             counts[OneShotResponseListener],
         )
 
-    def frame_received(self, frame: GeneralFrame) -> None:
+    def frame_received(self, frame: GeneralFrame) -> bool:
         """
-        Called when a frame has been received.
+        Called when a frame has been received. Returns whether or not the frame was
+        handled by any listener.
 
         XXX: Can be called multiple times in a single event loop step!
         """
@@ -321,28 +359,26 @@ class ZNP:
                 continue
 
             if not listener.resolve(command):
-                LOGGER.log(TRACE, "%s does not match %s", command, listener)
+                LOGGER.log(log.TRACE, "%s does not match %s", command, listener)
                 continue
 
             matched = True
-            LOGGER.log(TRACE, "%s matches %s", command, listener)
+            LOGGER.log(log.TRACE, "%s matches %s", command, listener)
 
             if isinstance(listener, OneShotResponseListener):
                 one_shot_matched = True
 
         if not matched:
-            LOGGER.warning("Received an unhandled command: %s", command)
+            self._unhandled_command(command)
 
-    async def iterator_for_responses(
-        self, responses
-    ) -> typing.AsyncGenerator[t.CommandBase, None]:
+        return matched
+
+    def _unhandled_command(self, command: t.CommandBase):
         """
-        Yields all matching responses as long as the async iterator is active.
+        Called when a command that is not handled by any listener is received.
         """
 
-        async with self.capture_responses(responses) as queue:
-            while True:
-                yield await queue.get()
+        LOGGER.warning("Received an unhandled command: %s", command)
 
     @contextlib.asynccontextmanager
     async def capture_responses(self, responses):
@@ -358,6 +394,19 @@ class ZNP:
         finally:
             self.remove_listener(listener)
 
+    @contextlib.asynccontextmanager
+    async def capture_responses_once(self, responses):
+        """
+        Captures all matched responses in a queue within the context manager.
+        """
+
+        future, listener = self.wait_for_responses(responses, context=True)
+
+        try:
+            yield future
+        finally:
+            self.remove_listener(listener)
+
     def callback_for_responses(self, responses, callback) -> CallbackResponseListener:
         """
         Creates a callback listener that matches any of the provided responses.
@@ -368,7 +417,7 @@ class ZNP:
 
         listener = CallbackResponseListener(responses, callback=callback)
 
-        LOGGER.log(TRACE, "Creating callback %s", listener)
+        LOGGER.log(log.TRACE, "Creating callback %s", listener)
 
         for header in listener.matching_headers():
             self._listeners[header].append(listener)
@@ -384,14 +433,14 @@ class ZNP:
 
         return self.callback_for_responses([response], callback)
 
-    def wait_for_responses(self, responses) -> asyncio.Future:
+    def wait_for_responses(self, responses, *, context=False) -> asyncio.Future:
         """
         Creates a one-shot listener that matches any *one* of the given responses.
         """
 
         listener = OneShotResponseListener(responses)
 
-        LOGGER.log(TRACE, "Creating one-shot listener %s", listener)
+        LOGGER.log(log.TRACE, "Creating one-shot listener %s", listener)
 
         for header in listener.matching_headers():
             self._listeners[header].append(listener)
@@ -399,7 +448,10 @@ class ZNP:
         # Remove the listener when the future is done, not only when it gets a result
         listener.future.add_done_callback(lambda _: self.remove_listener(listener))
 
-        return listener.future
+        if context:
+            return listener.future, listener
+        else:
+            return listener.future
 
     def wait_for_response(self, response: t.CommandBase) -> asyncio.Future:
         """
@@ -457,6 +509,7 @@ class ZNP:
                     ),
                 ]
             )
+
             self._uart.send(request.to_frame())
 
             # We should get a SRSP in a reasonable amount of time
@@ -479,7 +532,9 @@ class ZNP:
 
         return response
 
-    async def request_callback_rsp(self, *, request, callback, **response_params):
+    async def request_callback_rsp(
+        self, *, request, callback, timeout=None, **response_params
+    ):
         """
         Sends an SREQ, gets its SRSP confirmation, and waits for its real AREQ response.
         A bug-free version of:
@@ -491,29 +546,39 @@ class ZNP:
         from the UART and be handled in the same event loop step by ZNP.
         """
 
-        callback_response = self.wait_for_response(callback)
-        response = self.request(request, **response_params)
+        if timeout is None:
+            timeout = self._config[conf.CONF_ZNP_CONFIG][conf.CONF_ARSP_TIMEOUT]
 
-        await response
+        # The async context manager allows us to clean up resources upon cancellation
+        async with self.capture_responses_once([callback]) as callback_rsp:
+            await self.request(request, **response_params)
 
-        async with async_timeout.timeout(
-            self._config[conf.CONF_ZNP_CONFIG][conf.CONF_ARSP_TIMEOUT]
-        ):
-            return await callback_response
+            async with async_timeout.timeout(timeout):
+                return await callback_rsp
 
-    async def nvram_write(
-        self, nv_id: nvids.BaseNvIds, value, *, offset: t.uint8_t = 0
-    ):
+    async def nvram_delete(self, nv_id: nvids.BaseNvIds) -> bool:
         """
-        Convenience function for writing a value to NVRAM. Serializes all serializable
-        values and passes bytes directly.
+        Deletes an item from NVRAM. Returns whether or not the item existed.
         """
 
-        if not isinstance(nv_id, nvids.BaseNvIds):
-            raise ValueError(
-                "The nv_id param must be an instance of BaseNvIds. "
-                "Extend one of the tables in zigpy_znp.types.nvids."
-            )
+        length = (await self.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen
+
+        if length == 0:
+            return False
+
+        delete_rsp = await self.request(
+            c.SYS.OSALNVDelete.Req(Id=nv_id, ItemLen=length)
+        )
+
+        return delete_rsp.Status == t.Status.SUCCESS
+
+    async def nvram_write(self, nv_id: nvids.BaseNvIds, value, *, create: bool = False):
+        """
+        Writes a complete value to NVRAM, optionally resizing and creating the item if
+        necessary.
+
+        Serializes all serializable values and passes bytes directly.
+        """
 
         if hasattr(value, "serialize"):
             value = value.serialize()
@@ -523,22 +588,107 @@ class ZNP:
                 f" Got {nv_id!r}={value!r} (type {type(value)})"
             )
 
-        return await self.request(
-            c.SYS.OSALNVWrite.Req(Id=nv_id, Offset=offset, Value=t.ShortBytes(value)),
-            RspStatus=t.Status.SUCCESS,
-        )
+        if not value:
+            raise ValueError(f"NV item {nv_id!r} cannot be empty")
 
-    async def nvram_read(
-        self, nv_id: nvids.BaseNvIds, *, offset: t.uint8_t = 0
-    ) -> bytes:
+        length = (await self.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen
+
+        # Recreate the item if the length is not correct
+
+        # XXX: Some NVIDs don't really exist and Z-Stack doesn't behave consistently
+        #      when operations are performed on them.
+        if length != len(value) and nv_id != t.nvids.NwkNvIds.POLL_RATE_OLD16:
+            if not create:
+                if length == 0:
+                    raise KeyError(f"NV item does not exist: {nv_id!r}")
+                else:
+                    raise ValueError(
+                        f"Stored length and actual length differ:"
+                        f" {length} != {len(value)}"
+                    )
+
+            if length != 0:
+                await self.request(
+                    c.SYS.OSALNVDelete.Req(Id=nv_id, ItemLen=length),
+                    RspStatus=t.Status.SUCCESS,
+                )
+
+            await self.request(
+                c.SYS.OSALNVItemInit.Req(
+                    Id=nv_id,
+                    ItemLen=len(value),
+                    Value=t.ShortBytes(value[:244]),
+                ),
+                RspStatus=t.Status.NV_ITEM_UNINIT,
+            )
+
+        # 244 bytes is the most you can fit in a single `SYS.OSALNVWriteExt` command
+        for offset in range(0, len(value), 244):
+            await self.request(
+                c.SYS.OSALNVWriteExt.Req(
+                    Id=nv_id,
+                    Offset=offset,
+                    Value=t.ShortBytes(value[offset : offset + 244]),
+                ),
+                RspStatus=t.Status.SUCCESS,
+            )
+
+    async def nvram_read(self, nv_id: nvids.BaseNvIds) -> bytes:
         """
-        Reads a value from NVRAM.
+        Reads a complete value from NVRAM.
 
         Raises an `InvalidCommandResponse` error if the NVID doesn't exist.
         """
 
-        response = await self.request(
-            c.SYS.OSALNVRead.Req(Id=nv_id, Offset=offset), RspStatus=t.Status.SUCCESS,
-        )
+        # XXX: Some NVIDs don't really exist and Z-Stack behaves strangely with them
+        if nv_id == t.nvids.NwkNvIds.POLL_RATE_OLD16:
+            read_rsp = await self.request(
+                c.SYS.OSALNVRead.Req(Id=nv_id, Offset=0),
+                RspStatus=t.Status.SUCCESS,
+            )
 
-        return response.Value
+            return read_rsp.Value
+
+        # Every item has a length, even missing ones
+        length = (await self.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen
+
+        if length == 0:
+            raise KeyError(f"NV item does not exist: {nv_id!r}")
+
+        data = b""
+
+        # Not all items can be read out due to security policies, though this can easily
+        # be bypassed for most
+        if (
+            nvids.is_secure_nvid(nv_id)
+            and self._capabilities & t.MTCapabilities.CAP_SAPI
+        ):
+            # The SAPI "ConfigId" is only 8 bits, which means some nvids are not
+            # able to read this way
+            if nv_id > 0xFF:
+                raise SecurityError(
+                    f"NV item cannot be read due to security constraints: {nv_id!r}"
+                )
+
+            # It is impossible to read an item that will not fit in a single response
+            assert length <= 247
+
+            read_rsp = await self.request(
+                c.SAPI.ZBReadConfiguration.Req(ConfigId=nv_id),
+                RspStatus=t.Status.SUCCESS,
+                RspConfigId=nv_id,
+            )
+
+            data = read_rsp.Value
+        else:
+            while len(data) < length:
+                read_rsp = await self.request(
+                    c.SYS.OSALNVReadExt.Req(Id=nv_id, Offset=len(data)),
+                    RspStatus=t.Status.SUCCESS,
+                )
+
+                data += read_rsp.Value
+
+        assert len(data) == length
+
+        return data
