@@ -61,7 +61,7 @@ class BDBCommissioningMode(t.enum_flag_uint8):
 
 class InstallCodeFormat(t.enum_uint8):
     InstallCodeAndCRC = 0x01
-    KeyDeriveFromInstallCode = 0x02
+    KeyDerivedFromInstallCode = 0x02
 
 
 class AppConfig(t.CommandsBase, subsystem=t.Subsystem.APPConfig):
@@ -104,7 +104,9 @@ class AppConfig(t.CommandsBase, subsystem=t.Subsystem.APPConfig):
         0x03,
         req_schema=(
             t.Param(
-                "AllowRejoin", t.Bool, "whether or not the Trust center allows rejoins",
+                "AllowRejoin",
+                t.Bool,
+                "whether or not the Trust center allows rejoins",
             ),
         ),
         rsp_schema=t.STATUS_SCHEMA,
