@@ -1,11 +1,11 @@
 import logging
 import argparse
+
 import coloredlogs
 
-from zigpy_znp.logger import _TRACE
+import zigpy_znp.logger as log
 
-
-LOG_LEVELS = [logging.INFO, logging.DEBUG, _TRACE]
+LOG_LEVELS = [logging.INFO, logging.DEBUG, log._TRACE]
 
 
 class CustomArgumentParser(argparse.ArgumentParser):
@@ -13,11 +13,23 @@ class CustomArgumentParser(argparse.ArgumentParser):
         args = super().parse_args(args, namespace)
 
         # Since we're running as a CLI tool, install our own log level and color logger
+        log.TRACE = log._TRACE
+        logging.addLevelName(log.TRACE, "TRACE")
+
+        # But still allow the user to configure verbosity
         verbosity = args.verbosity
         log_level = LOG_LEVELS[min(max(0, verbosity), len(LOG_LEVELS) - 1)]
 
-        logging.addLevelName(_TRACE, "TRACE")
-        coloredlogs.install(level=log_level)
+        # coloredlogs uses "spam" for level 5, not "trace"
+        level_styles = coloredlogs.DEFAULT_LEVEL_STYLES.copy()
+        level_styles["trace"] = level_styles["spam"]
+
+        logging.getLogger().setLevel(log_level)
+
+        coloredlogs.install(
+            level=log_level,
+            level_styles=level_styles,
+        )
 
         return args
 
