@@ -2,19 +2,16 @@ import typing
 import asyncio
 import logging
 import warnings
+from collections import defaultdict
 
 import serial
 import serial.tools
-
-from collections import defaultdict
-
 from serial.tools.list_ports import comports as list_com_ports
 
 import zigpy_znp.config as conf
 import zigpy_znp.frames as frames
-
+import zigpy_znp.logger as log
 from zigpy_znp.types import Bytes
-from zigpy_znp.logger import TRACE
 from zigpy_znp.exceptions import InvalidFrame
 
 with warnings.catch_warnings():
@@ -26,7 +23,9 @@ with warnings.catch_warnings():
     )
     import serial_asyncio  # noqa: E402
 
+
 LOGGER = logging.getLogger(__name__)
+RTS_TOGGLE_DELAY = 0.15  # seconds
 
 
 class BufferTooShort(Exception):
@@ -40,30 +39,25 @@ class ZnpMtProtocol(asyncio.Protocol):
         self._transport = None
         self._connected_event = asyncio.Event()
 
-    @property
-    def transport(self) -> typing.Optional[serial_asyncio.SerialTransport]:
-        """Return current transport."""
-        return self._transport
-
     def close(self) -> None:
         """Closes the port."""
         self._buffer.clear()
-        self.transport.close()
+
+        if self._transport is not None:
+            self._transport.close()
+            self._transport = None
 
     def connection_lost(self, exc: typing.Optional[Exception]) -> None:
         """Connection lost."""
 
-        self._buffer.clear()
-
         if exc is not None:
-            LOGGER.warning(
-                "Lost connection to %s", self._transport.serial.name, exc_info=exc
-            )
+            LOGGER.warning("Lost connection", exc_info=exc)
 
+        LOGGER.debug("Closing serial port")
+
+        self.close()
         self._api.connection_lost(exc)
 
-        LOGGER.debug("Closing %s serial port", self._transport.serial.name)
-
     def connection_made(self, transport: serial_asyncio.SerialTransport) -> None:
         """Opened serial port."""
         self._transport = transport
@@ -71,14 +65,16 @@ class ZnpMtProtocol(asyncio.Protocol):
 
         self._connected_event.set()
 
+        self._api.connection_made()
+
     def data_received(self, data: bytes) -> None:
         """Callback when data is received."""
         self._buffer += data
 
-        LOGGER.log(TRACE, "Received data: %s", Bytes.__repr__(data))
+        LOGGER.log(log.TRACE, "Received data: %s", Bytes.__repr__(data))
 
         for frame in self._extract_frames():
-            LOGGER.log(TRACE, "Parsed frame: %s", frame)
+            LOGGER.log(log.TRACE, "Parsed frame: %s", frame)
 
             try:
                 self._api.frame_received(frame.payload)
@@ -92,8 +88,8 @@ class ZnpMtProtocol(asyncio.Protocol):
         self._transport_write(frames.TransportFrame(payload).serialize())
 
     def _transport_write(self, data: bytes) -> None:
-        LOGGER.log(TRACE, "Sending data: %s", Bytes.__repr__(data))
-        self.transport.write(data)
+        LOGGER.log(log.TRACE, "Sending data: %s", Bytes.__repr__(data))
+        self._transport.write(data)
 
     def _extract_frames(self) -> typing.Iterator[frames.TransportFrame]:
         """Extracts frames from the buffer until it is exhausted."""
@@ -204,14 +200,15 @@ def guess_port() -> str:
 
     if len(candidates) > 1:
         LOGGER.warning(
-            "Found multiple possible Texas Instruments devices: %s", candidates,
+            "Found multiple possible Texas Instruments devices: %s",
+            candidates,
         )
         LOGGER.warning("Picking the first one: %s", port)
 
     return port.device
 
 
-async def connect(config: conf.ConfigType, api) -> ZnpMtProtocol:
+async def connect(config: conf.ConfigType, api, *, toggle_rts=True) -> ZnpMtProtocol:
     loop = asyncio.get_running_loop()
 
     port = config[conf.CONF_DEVICE_PATH]
@@ -221,27 +218,39 @@ async def connect(config: conf.ConfigType, api) -> ZnpMtProtocol:
     if port == "auto":
         port = guess_port()
 
-    xonxoff, rtscts = {
-        None: (False, False),
-        "hardware": (False, True),
-        "software": (True, False),
-    }[flow_control]
-
     LOGGER.debug("Connecting to %s at %s baud", port, baudrate)
 
     transport, protocol = await serial_asyncio.create_serial_connection(
-        loop,
-        lambda: ZnpMtProtocol(api),
+        loop=loop,
+        protocol_factory=lambda: ZnpMtProtocol(api),
         url=port,
         baudrate=baudrate,
         parity=serial.PARITY_NONE,
         stopbits=serial.STOPBITS_ONE,
-        xonxoff=xonxoff,
-        rtscts=rtscts,
+        xonxoff=(flow_control == "software"),
+        rtscts=(flow_control == "hardware"),
     )
 
     await protocol._connected_event.wait()
 
+    # Skips the bootloader on slaesh's CC2652R USB stick
+    if toggle_rts:
+        LOGGER.debug("Toggling RTS/CTS to skip CC2652R bootloader")
+        transport.serial.dtr = False
+        transport.serial.rts = False
+
+        await asyncio.sleep(RTS_TOGGLE_DELAY)
+
+        transport.serial.dtr = False
+        transport.serial.rts = True
+
+        await asyncio.sleep(RTS_TOGGLE_DELAY)
+
+        transport.serial.dtr = False
+        transport.serial.rts = False
+
+        await asyncio.sleep(RTS_TOGGLE_DELAY)
+
     LOGGER.debug("Connected to %s at %s baud", port, baudrate)
 
     return protocol
